package ru.dbotthepony.mc.otm.block.entity.decorative

import com.google.common.collect.ImmutableList
import it.unimi.dsi.fastutil.ints.Int2IntArrayMap
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap
import it.unimi.dsi.fastutil.objects.Object2IntMap
import net.minecraft.core.BlockPos
import net.minecraft.core.HolderLookup
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Fluids
import net.neoforged.neoforge.capabilities.Capabilities
import net.neoforged.neoforge.fluids.FluidStack
import net.neoforged.neoforge.fluids.capability.IFluidHandler
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.container.slotted.ContainerSlot
import ru.dbotthepony.mc.otm.container.slotted.SlottedContainer
import ru.dbotthepony.kommons.guava.immutableList
import ru.dbotthepony.kommons.guava.immutableMap
import ru.dbotthepony.mc.otm.util.isNotEmpty
import ru.dbotthepony.mc.otm.util.mapPresent
import ru.dbotthepony.mc.otm.util.set
import ru.dbotthepony.mc.otm.menu.decorative.PainterMenu
import ru.dbotthepony.mc.otm.registry.game.MBlockEntities
import java.util.*

class PainterBlockEntity(blockPos: BlockPos, blockState: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.PAINTER, blockPos, blockState), IFluidHandler {
	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return PainterMenu(containerID, inventory, this)
	}

	private inner class InputSlot(container: SlottedContainer, slot: Int) : ContainerSlot(container, slot) {
		override fun canAutomationPlaceItem(itemStack: ItemStack): Boolean {
			if (!super.canAutomationPlaceItem(itemStack))
				return false

			if (waterStored() < MAX_WATER_STORAGE) {
				itemStack.getCapability(Capabilities.FluidHandler.ITEM)?.let {
					val drain = it.drain(FluidStack(Fluids.WATER, MAX_WATER_STORAGE - waterStored()), IFluidHandler.FluidAction.SIMULATE)

					if (drain.isNotEmpty) {
						return true
					}
				}
			}

			val dye = DyeColor.getColor(itemStack) ?: return false
			return dyeStored(dye) + HUE_PER_ITEM <= MAX_STORAGE
		}

		override fun canAutomationTakeItem(desired: Int): Boolean {
			return false
		}

		override fun modifyAutomationPlaceCount(itemStack: ItemStack): Int {
			if (itemStack.getCapability(Capabilities.FluidHandler.ITEM) != null)
				return 1

			val dye = DyeColor.getColor(itemStack) ?: return 0
			return itemStack.count.coerceAtMost((MAX_STORAGE - dyeStored(dye)) / HUE_PER_ITEM - item.count)
		}
	}

	val dyeInput = SlottedContainer.simple(1, ::InputSlot, ::markDirtyFast)
	// null - water
	val dyeStored = Object2IntArrayMap<DyeColor?>()

	var isBulk = false
		set(value) {
			field = value
			markDirtyFast()
		}

	override fun getTanks(): Int {
		return 1
	}

	override fun getFluidInTank(tank: Int): FluidStack {
		if (tank != 1) return FluidStack.EMPTY
		if (waterStored() == 0) return FluidStack.EMPTY
		return FluidStack(Fluids.WATER, waterStored())
	}

	override fun getTankCapacity(tank: Int): Int {
		if (tank == 1) return MAX_WATER_STORAGE
		return 0
	}

	override fun isFluidValid(tank: Int, stack: FluidStack): Boolean {
		return tank == 0 && stack.fluid == Fluids.WATER
	}

	override fun fill(resource: FluidStack, action: IFluidHandler.FluidAction): Int {
		if (resource.amount <= 0 || resource.fluid != Fluids.WATER || waterStored() >= MAX_WATER_STORAGE) return 0
		val diff = ((waterStored().toLong() + resource.amount.toLong()).coerceAtMost(MAX_WATER_STORAGE.toLong()) - waterStored().toLong()).toInt()
		if (action.simulate() || diff == 0) return diff
		dyeStored[null] = waterStored() + diff
		return diff
	}

	override fun drain(resource: FluidStack, action: IFluidHandler.FluidAction): FluidStack {
		return FluidStack.EMPTY
	}

	override fun drain(maxDrain: Int, action: IFluidHandler.FluidAction): FluidStack {
		return FluidStack.EMPTY
	}

	init {
		addDroppableContainer(dyeInput)
		savetables.stateful(dyeInput, INVENTORY_KEY)
		savetablesConfig.bool(::isBulk)
		exposeGlobally(Capabilities.FluidHandler.BLOCK, this)
	}

	fun takeDyes(dyes: Map<out DyeColor?, Int>) {
		for ((dye, amount) in dyes) {
			if (dye == null) {
				for (i in 0 until amount) {
					if (waterStored() >= 1) {
						dyeStored[dye] = waterStored() - 1
					}
				}
			} else {
				for (i in 0 until amount) {
					mixer(dye)?.mix(dyeStored)

					if (dyeStored(dye) > 0) {
						dyeStored[dye] = dyeStored(dye) - 1
					}
				}
			}
		}

		markDirtyFast()
	}

	val config = ConfigurableItemHandler(input = dyeInput)

	fun waterStored(): Int {
		return dyeStored.getInt(null)
	}

	fun dyeStored(dye: DyeColor?): Int {
		return dyeStored.getInt(dye)
	}

	override fun saveShared(nbt: CompoundTag, registry: HolderLookup.Provider) {
		super.saveShared(nbt, registry)

		nbt["dyes"] = CompoundTag().also {
			for ((k, v) in dyeStored) {
				it[k?.getName() ?: "water"] = v
			}
		}
	}

	override fun loadAdditional(nbt: CompoundTag, registry: HolderLookup.Provider) {
		super.loadAdditional(nbt, registry)

		dyeStored.clear()

		nbt.mapPresent("dyes") { it: CompoundTag ->
			for (k in it.allKeys) {
				if (k == "water")
					dyeStored[null] = it.getInt("water")
				else
					dyeStored[DyeColor.entries.firstOrNull { it.getName() == k } ?: continue] = it.getInt(k)
			}
		}
	}

	override fun tick() {
		super.tick()

		for (slot in dyeInput.slotIterator()) {
			if (waterStored() < MAX_WATER_STORAGE) {
				slot.item.getCapability(Capabilities.FluidHandler.ITEM)?.let {
					val drain = it.drain(FluidStack(Fluids.WATER, MAX_WATER_STORAGE - waterStored()), IFluidHandler.FluidAction.EXECUTE)

					if (drain.isNotEmpty) {
						dyeStored[null] = waterStored() + drain.amount
						slot.item = it.container
					}
				}
			}

			val dye = DyeColor.entries.firstOrNull { slot.item.`is`(it.tag) } ?: continue
			val stored = dyeStored(dye)

			if (stored + HUE_PER_ITEM <= MAX_STORAGE) {
				slot.item.shrink(1)
				slot.setChanged()
				dyeStored[dye] = stored + HUE_PER_ITEM
			}
		}
	}

	class Mixer(val color: DyeColor, vararg mixing: ImmutableList<Optional<DyeColor>>) : Map.Entry<DyeColor, Mixer> {
		val mixing: ImmutableList<ImmutableList<Optional<DyeColor>>> = ImmutableList.copyOf(mixing)

		override val key: DyeColor
			get() = color
		override val value: Mixer
			get() = this

		private fun mix(input: Object2IntMap<DyeColor?>, seen: MutableSet<DyeColor>, stack: MutableSet<DyeColor>) {
			if (input.getInt(color) > 0 || color in seen || color in stack) return
			stack.add(color)

			for (ingredients in mixing) {
				val copy = Object2IntArrayMap(input)
				var i = 0
				val volumes = Int2IntArrayMap()

				// страховка от взаимной блокировки когда последующий ингредиент требует предыдущий.
				// создаёт горку из значений
				for (i2 in ingredients.indices)
					volumes[i2] = 1

				while (!ingredients.all { copy.getInt(it.orElse(null)) > 0 } && i++ < 20) {
					ingredients.withIndex().forEach {
						for (t in 0 until volumes[it.index])
							it.value.ifPresent { mixer(it)?.mix(copy, seen, stack) }
					}

					var increase = i
					val iter = ingredients.indices.iterator()

					while (iter.hasNext() && increase-- > 0) {
						volumes[iter.nextInt()]++
					}
				}

				if (ingredients.all { copy.getInt(it.orElse(null)) > 0 }) {
					ingredients.forEach { copy[it.orElse(null)] = copy.getInt(it.orElse(null)) - 1 }
					copy[color] = ingredients.size

					input.putAll(copy)
					stack.remove(color)
					return
				}
			}

			stack.remove(color)
			seen.add(color)
		}

		fun mix(input: Object2IntMap<DyeColor?>) {
			mix(input, EnumSet.noneOf(DyeColor::class.java), EnumSet.noneOf(DyeColor::class.java))
		}

		fun isAvailable(input: Object2IntMap<DyeColor?>): Boolean {
			if (input.getInt(color) > 0) return true
			return Object2IntArrayMap(input).also(::mix).getInt(color) > 0
		}
	}

	companion object {
		fun mixer(color: DyeColor): Mixer? {
			return MIXING[color]
		}

		val MIXING = immutableMap<DyeColor, Mixer> {
			put(Mixer(DyeColor.BLACK, immutableList(Optional.of(DyeColor.CYAN), Optional.of(DyeColor.MAGENTA), Optional.of(DyeColor.YELLOW))))
			put(Mixer(DyeColor.RED, immutableList(Optional.of(DyeColor.MAGENTA), Optional.of(DyeColor.YELLOW))))
			put(Mixer(DyeColor.GREEN, immutableList(Optional.of(DyeColor.CYAN), Optional.of(DyeColor.YELLOW))))
			put(Mixer(DyeColor.BLUE, immutableList(Optional.of(DyeColor.CYAN), Optional.of(DyeColor.MAGENTA))))
			put(Mixer(DyeColor.LIGHT_BLUE, immutableList(Optional.of(DyeColor.WHITE), Optional.of(DyeColor.BLUE))))
			put(Mixer(DyeColor.LIME, immutableList(Optional.of(DyeColor.WHITE), Optional.of(DyeColor.GREEN))))
			put(Mixer(DyeColor.ORANGE, immutableList(Optional.of(DyeColor.WHITE), Optional.of(DyeColor.RED), Optional.empty())))
			put(Mixer(DyeColor.PURPLE, immutableList(Optional.of(DyeColor.CYAN), Optional.of(DyeColor.MAGENTA), Optional.empty())))
			put(Mixer(DyeColor.BROWN, immutableList(Optional.of(DyeColor.YELLOW), Optional.of(DyeColor.MAGENTA), Optional.of(DyeColor.BLACK), Optional.empty())))
			put(Mixer(DyeColor.PINK, immutableList(Optional.of(DyeColor.YELLOW), Optional.of(DyeColor.MAGENTA), Optional.of(DyeColor.WHITE), Optional.empty())))
			put(Mixer(DyeColor.GRAY, immutableList(Optional.of(DyeColor.WHITE), Optional.of(DyeColor.BLACK), Optional.empty())))
			put(Mixer(DyeColor.LIGHT_GRAY, immutableList(Optional.of(DyeColor.WHITE), Optional.of(DyeColor.WHITE), Optional.of(DyeColor.BLACK), Optional.empty())))
		}

		const val MAX_STORAGE = 256
		const val HUE_PER_ITEM = 32
		const val MAX_WATER_STORAGE = 4_000
	}
}
