package ru.dbotthepony.mc.otm.util

import net.minecraft.core.BlockPos
import net.minecraft.core.SectionPos
import net.minecraft.util.RandomSource
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.Level
import net.minecraft.world.level.LevelAccessor
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.Blocks
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.chunk.ChunkSource
import net.minecraft.world.level.chunk.LevelChunk

fun LevelAccessor.getBlockStateNow(pos: BlockPos): BlockState {
	return getBlockStateNowOrNull(pos) ?: Blocks.AIR.defaultBlockState()
}

fun LevelAccessor.getBlockStateNowOrNull(pos: BlockPos): BlockState? {
	return chunkSource.getChunkNow(
		SectionPos.blockToSectionCoord(pos.x),
		SectionPos.blockToSectionCoord(pos.z)
	)?.getBlockState(pos)
}

fun LevelAccessor.getBlockEntityNow(pos: BlockPos): BlockEntity? {
	return chunkSource.getChunkNow(
		SectionPos.blockToSectionCoord(pos.x),
		SectionPos.blockToSectionCoord(pos.z)
	)?.getBlockEntity(pos)
}

fun ChunkSource.getChunkNow(pos: BlockPos): LevelChunk? {
	return getChunkNow(SectionPos.blockToSectionCoord(pos.x), SectionPos.blockToSectionCoord(pos.z))
}

fun ChunkSource.getChunkNow(pos: ChunkPos): LevelChunk? {
	return getChunkNow(pos.x, pos.z)
}

interface IMatteryLevel {
	val otmRandom: RandomSource?
}

/**
 * OTM provided [RandomSource], which has better statistical parameters
 *
 * Original Minecraft use LCG, which may show bad behavior when repeatedly sampled *a lot*,
 * which is what [Level]'s random is used for. OTM provided PRNG should behave better in this scenario.
 *
 * The way OTM uses random generator in its code will quickly cause LCG used in Minecraft to show its bias
 * because LCG in minecraft samples its highest 48 bits, which gives us at best 2^16 period in the lowest bit returned by LCG.
 * Which it doesn't sound bad, it quickly causes RNG become biased the quicker/more it is sampled on each tick, especially considering
 * some may use `level.random.nextInt(chance) == 0` to determine chance of something happening,
 * which will get extremely biased on heavy RNG congested environment
 * If we avoid sampling Level's generator this much, we won't suffer from bias in our own code, as well as avoid biasing other mods this much.
 *
 * The "2^16 period" problem is also might be the reason why Entities get their own instance of RandomSource,
 * and Mob Goals use random exactly the way described above (`nextInt(chance)`), which can and will suffer
 * from bias the moment mob exists in world for more than 2^16 ticks (but actual bias will happen sooner
 * because RNG is not sampled only once per tick, obviously)
 */
val Level.otmRandom: RandomSource get() = (this as IMatteryLevel).otmRandom ?: random

fun Level.gracefulBlockBreak(blockPos: BlockPos, block: BlockState = getBlockState(blockPos)) {
	val tile = getBlockEntity(blockPos)
	val state = block.fluidState.createLegacyBlock()
	block.onRemove(this, blockPos, state, false)
	Block.dropResources(block, this, blockPos, tile)
	block.block.destroy(this, blockPos, block)
	setBlock(blockPos, state, Block.UPDATE_ALL)
}
